# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class SchweizerSklarTNorm(BaseTNorm):
    def __init__(self,
                 p: Optional[Tensor] = None,
                 default_p: float = 0.1) -> None:
        super().__init__()
        self.p = p
        if self.p is None:
            self.p = nn.Parameter(torch.tensor(default_p))
        assert len(self.p.shape) == 0

        self.relu = torch.nn.ReLU()
        self.eps = 0.001

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        res: Optional[Tensor] = None
        # Careful that the value of parametr p can be 10e-8, thus maybe add eps.



        # self.p.data = torch.clamp(self.p.data, min=-4)
        if torch.isnan(self.p):
            print("NAN a BRAT")
            self.p.data = torch.tensor(0.1)
        if self.p.item() < 0.0:
            res: Tensor = (a ** self.p + b ** self.p - 1.0) ** (1.0 / self.p)
        elif self.p.item() == 0.0:
            res: Tensor = a * b
        elif self.p.item() > 0.0:
            res: Tensor = (a ** self.p + b ** self.p - 1.0)
            res: Tensor = torch.maximum(res, torch.tensor(0.0))
            res: Tensor = res ** (1.0 / self.p)

        # if torch.isnan(res):
        #     print("NAN a BRAT")
        #     print(self.p)

        return res



# -*- coding: utf-8 -*-

import torch
from torch import Tensor

from torchnorms.tnorms.base import BaseTNorm


class MinimumTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return torch.min(a, b)


class ProductTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return a * b


class LukasiewiczTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return torch.relu(a + b - 1.0)


class DrasticTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        b_mask = b * (a >= 1.0)
        a_mask = a * (b >= 1.0)
        return torch.max(a_mask, b_mask)


class EinsteinTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return (a * b) / (1.0 + (1.0 - a) * (1.0 - b))


class BoundedTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return torch.maximum(torch.tensor(0), a + b - 1.0)


class NilpotentMinimumTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        mask = (a + b) > 1.0
        return torch.min(a, b) * mask


class HamacherProductTNorm(BaseTNorm):
    def __init__(self) -> None:
        super().__init__()

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        a_nnz = a > 0.0
        b_nnz = b > 0.0
        mask = torch.logical_and(a_nnz, b_nnz)
        res = (a * b) / (a + b - a * b)
        return res * mask



# -*- coding: utf-8 -*-

from torch import nn, Tensor

from abc import abstractmethod

import logging

logger = logging.getLogger(__name__)


class BaseTCoNorm(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def __call__(self,
              a: Tensor,
              b: Tensor) -> Tensor:
        raise NotImplementedError

    def norm(self,
               a: Tensor,
               b: Tensor) -> Tensor:
        return 1.0 - self(1.0 - a, 1.0 - b)



import torchnorms.tconorms.base
import torchnorms.tconorms.classic
import torchnorms.tconorms.dombi
import torchnorms.tconorms.dubois
import torchnorms.tconorms.hamacher
import torchnorms.tconorms.ss
import torchnorms.tconorms.weber
import torchnorms.tconorms.yager



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor

from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class DombiTNorm(BaseTNorm):
    def __init__(self,
                 p: Optional[Tensor] = None,
                 default_p: float = 0.1) -> None:
        super().__init__()
        self.p = p
        if self.p is None:
            self.p = nn.Parameter(torch.tensor(default_p))
        assert len(self.p.shape) == 0
        self.relu = torch.nn.ReLU()
        self.eps = 0.001

    def __call__(self,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        self.p.data = self.relu(self.p.data)

        if self.p == 0:
            # eps is added to avoid 0
            self.p.data += self.eps

        p_1 = torch.pow((1 - a) / a, self.p)
        p_2 = torch.pow((1 - b) / b, self.p)
        p_3 = 1.0 + torch.pow(p_1 + p_2, 1.0 / self.p)
        res = 1.0 / p_3

        return res



# -*- coding: utf-8 -*-

import torch
from torch import Tensor

from torchnorms.tconorms.base import BaseTCoNorm


class MinimumCoNorm(BaseTCoNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return torch.maximum(a, b)


class ProductTCoNorm(BaseTCoNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return 1 - ((1 - a) * (1 - b))


class LukasiewiczTCoNorm(BaseTCoNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return torch.minimum(a+b,torch.tensor(1)) # Carefull not to overbroadcast


class EinsteinTCoNorm(BaseTCoNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return (a + b) / (1.0 + (a * b))

class BoundedTCoNorm(BaseTCoNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        return torch.minimum(torch.tensor(1), a + b)


class NilpotentMinimumTCoNorm(BaseTCoNorm):
    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def __call__(cls,
                 a: Tensor,
                 b: Tensor) -> Tensor:
        mask_1 = a + b < 1.0
        mask_2 = a + b >= 0
        return mask_2 + torch.max(a, b) * mask_1
